@@ -11,7 +11,7 @@ our ($VERSION);
 use XSLoader ();
 
 BEGIN {
-    $VERSION = '0.23_01';
+    $VERSION = '0.30_01';
     XSLoader::load();
 }
 
@@ -64,6 +64,48 @@ The encoder currently does not handle any other blessed references.
 =item * C<canonical> - A boolean that makes the encoder output
 CBOR in L<canonical form|https://tools.ietf.org/html/rfc7049#section-3.9>.
 
+=item * C<string_encode_mode> - Takes one of:
+
+=over
+
+=item * C<sv>: The default mode of operation. If the string’s internal
+UTF8 flag is set, it will become a CBOR text string; otherwise, it will be
+CBOR binary. This is good for IPC with other Perl code but isn’t a very
+friendly default for working with other languages that probably expect more
+reliably-typed strings.
+
+This configuration is B<NOT> recommended; it’s the default behavior because
+it’s the only configuration that can reasonably fulfill that role. This is
+also the only way to output text and binary strings in a single CBOR document.
+
+=item * C<encode_text>: Treats all strings as unencoded characters.
+All CBOR strings will be text.
+
+This is probably what you want if you’re
+following the receive-decode-process-encode-output workflow that
+L<perlunitut> recommends (which you might be doing via C<use utf8>)
+B<AND> if you intend for your CBOR to contain exclusively text.
+
+(Perl internals note: if !SvUTF8, the CBOR will be the UTF8-upgraded
+version.)
+
+=item * C<as_text>: Treats all strings as octets of UTF-8.
+Wide characters are thus invalid input. All CBOR strings will be text.
+
+This is probably what you want if you forgo character decoding (and encoding),
+treating all input as octets, B<BUT> you still intend for your CBOR to
+contain exclusively text.
+
+(Perl internals note: if SvUTF8, the CBOR will be the downgraded version.)
+
+=item * C<as_binary>: Like C<as_text>, but outputs CBOR binary
+instead of text.
+
+This is probably what you want if your application is “all binary,
+all the time”.
+
+=back
+
 =item * C<text_keys> - EXPERIMENTAL. Encodes all Perl hash keys as CBOR text.
 If you use this mode then your strings B<must> be properly decoded, or else
 the output CBOR may mangle your string.
@@ -101,7 +143,8 @@ Notes on mapping Perl to CBOR:
 integer, float, string, etc.) determines its CBOR encoding.
 
 =item * Perl doesn’t currently provide reliable binary/character string types.
-CBOR::Free tries to distinguish anyway by looking at a string’s UTF8 flag: if
+CBOR::Free, in its default configuration, tries to distinguish anyway by
+looking at a string’s UTF8 flag: if
 set, then the string becomes CBOR text; otherwise, it’ll be CBOR binary.
 That’s not always going to work, though. A trivial example:
 
@@ -119,8 +162,9 @@ The crucial point, though, is that, because Perl itself doesn’t guarantee
 the reliable string types that CBOR recognizes, any heuristics we apply
 to distinguish one from the other are a “best-guess” merely.
 
-B<IMPORTANT:> Whatever consumes your Perl-sourced CBOR B<MUST> account
-for the prospect of an incorrectly-typed string.
+B<IMPORTANT:> If you use the default encoding configuration, whatever
+consumes your Perl-sourced CBOR B<MUST> account for the prospect of an
+incorrectly-typed string.
 
 =item * The above applies also to strings vs. numbers: whatever consumes
 your Perl-sourced CBOR B<MUST> account for the prospect of numbers that
@@ -157,14 +201,33 @@ Notes on mapping CBOR to Perl:
 
 =over
 
-=item * CBOR text strings become Perl strings with the internal UTF8 flag set.
-CBOR binary strings become Perl strings I<without> that flag set. This is
-a mostly-internal distinction in Perl that doesn’t actually constitute
-separate byte/character string types, but it’s at least something similar.
+=item * C<decode()> decodes CBOR text strings as UTF-8-decoded Perl strings.
+CBOR binary strings become undecoded Perl strings.
 
-Note that invalid UTF-8 in a CBOR text string is considered
+(See L<CBOR::Free::Decoder> and L<CBOR::Free::SequenceDecoder> for more
+character-decoding options.)
+
+CBOR::Free guarantees that its UTF-8-decoding will I<always> enable the
+resulting Perl string’s internal UTF8 flag—even for “invariant” strings
+like C<abc>. CBOR::Free also guarantees
+the inverse: an undecoded Perl string from CBOR::Free will I<never> have
+its internal UTF8 flag enabled.
+
+Notes:
+
+=over
+
+=item * Invalid UTF-8 in a CBOR text string is considered
 invalid input and will thus prompt a thrown exception.
 
+=item * You can reliably use C<utf8::is_utf8()> to determine if a given Perl
+string came from CBOR text or binary, but B<ONLY> if you test the scalar as
+it appears in the newly-decoded data structure itself. Ordinarily you
+should avoid C<is_utf8()>, but with CBOR::Free-created strings
+this limited use case is legitimate and potentially gainful.
+
+=back
+
 =item * The only map keys that C<decode()> accepts are integers and strings.
 An exception is thrown if the decoder finds anything else as a map key.
 Note that, because Perl does not distinguish between binary and text strings,
