@@ -24,6 +24,8 @@ decoder. This interface allows interpretation of tagged values.
 
 #----------------------------------------------------------------------
 
+use parent qw( CBOR::Free::Decoder::Base );
+
 use CBOR::Free ();
 
 #----------------------------------------------------------------------
@@ -36,8 +38,6 @@ Creates a new CBOR decoder object.
 
 =cut
 
-sub new { bless {} }    # TODO: implement in XS, and store a context.
-
 #----------------------------------------------------------------------
 
 =head2 $data = I<OBJ>->decode( $CBOR )
@@ -55,16 +55,51 @@ otherwise ignored.
 =head2 $enabled_yn = I<OBJ>->preserve_references( [$ENABLE] )
 
 Enables/disables recognition of CBOR’s shared references. (If no
-argument is given, shared references wil be enabled.)
+argument is given, shared references will be enabled.)
 
 B<HANDLE WITH CARE.> This option can cause CBOR::Free to create circular
 references, which can cause memory leaks if not handled properly.
 
 =cut
 
-sub preserve_references {
-    return $_[0]{'_preserve_references'} = (@_ > 1 ? !!$_[1] : 1);
-}
+#----------------------------------------------------------------------
+
+=head2 $enabled_yn = I<OBJ>->naive_utf8( [$ENABLE] )
+
+Same interface as C<preserve_references()>, but this option tells I<OBJ>
+to forgo UTF-8 validation of CBOR text strings when enabled. This speeds up
+decoding of text strings but may confuse Perl if invalid UTF-8 is given in
+a CBOR text string. That may or may not break your application.
+
+This I<should> be safe in contexts—such as IPC—where you control the CBOR
+serialization and can thus ensure validity of the encoded text.
+
+If in doubt, leave this off.
+
+=cut
+
+#----------------------------------------------------------------------
+
+=head2 $obj = I<OBJ>->string_decode_cbor();
+
+This causes I<OBJ> to decode strings according to their CBOR type:
+text strings are UTF8-decoded; binary strings are left as-is. This is
+the default configuration, à la C<CBOR::Free::decode()>.
+
+=head2 $obj = I<OBJ>->string_decode_never();
+
+This causes I<OBJ> to leave all strings undecoded. This is useful for
+applications that treat all strings as octet sequences. Note that CBOR
+text strings will still be validated as UTF-8 unless C<naive_utf8()> is
+enabled.
+
+=head2 $obj = I<OBJ>->string_decode_always();
+
+This causes I<OBJ> to decode all CBOR strings (including binary strings)
+as UTF-8, applying appropriate pre-validation unless C<naive_utf8()> is
+enabled. This is useful if you expect all strings (including binary) to be
+UTF-8 and want to handle them in Perl as character strings instead of
+byte strings.
 
 #----------------------------------------------------------------------
 
@@ -85,16 +120,4 @@ here will be ignored.
 
 =cut
 
-use constant _TAG_PACK_TMPL => eval { pack 'Q' } ? 'Q' : 'L';
-
-sub set_tag_handlers {
-    my ($self, @tag_cb) = @_;
-
-    while (my ($tag, $cb) = splice @tag_cb, 0, 2) {
-        $self->{'_tag_decode_callback'}{ pack( _TAG_PACK_TMPL(), $tag ) } = $cb;
-    }
-
-    return $self;
-}
-
 1;
